from dataclasses import dataclass
import struct
from enum import Enum

from obspy import UTCDateTime


class Unpacker:
    """
    A wrapper around struct.unpack() to unpack binary data without having to
    explicitly define the byte order in the format string. Also restrict the
    type of format to str and buffer to bytes.
    """
    def __init__(self, byte_order_char: str = '') -> None:
        self.byte_order_char = byte_order_char

    def unpack(self, format: str, buffer: bytes):
        """

        :param format:
        :param buffer: the byte string
        :return: a tuple containing the unpacked values.
        """
        default_byte_order_chars = ('@', '=', '>', '<', '!')
        if format.startswith(default_byte_order_chars):
            format = self.byte_order_char + format[:1]
        else:
            format = self.byte_order_char + format
        return struct.unpack(format, buffer)


@dataclass
class FixedHeader:
    sequence_number: bytes
    data_header_quality_indicator: bytes
    reserved: bytes
    station: bytes
    location: bytes
    channel: bytes
    net_code: bytes
    record_start_time: bytes
    sample_count: bytes
    sample_rate_factor: bytes
    sample_rate_multiplier: bytes
    activity_flags: bytes
    io_and_clock_flags: bytes
    data_quality_flags: bytes
    blockette_count: bytes
    time_correction: bytes
    data_offset: bytes
    first_blockette_offset: bytes


@dataclass
class Blockette1000:
    blockette_type: bytes
    next_blockette_offset: bytes
    encoding_format: bytes
    byte_order: bytes
    record_length: bytes
    reserved_byte: bytes


@dataclass
class RecordMetadata:
    station: str
    location: str
    channel: str
    network: str
    start_time: UTCDateTime
    sample_count: int
    sample_rate: float


class EncodingFormat(Enum):
    ASCII = 0
    INT_16_BIT = 1
    INT_24_BIT = 2
    INT_32_BIT = 3
    IEEE_FLOAT_32_BIT = 4
    IEEE_FLOAT_64_BIT = 5
    STEIM_1 = 10
    STEIM_2 = 11


def get_header_endianness(header: FixedHeader):
    """
    Determine the endianness of the fixed header of a data record. Works by
    checking if the decoded record start time has a sane value if the header
    is assumed to be big-endian.

    WARNING: This check fails on three dates: 2056-1, 2056-256, and 2056-257.
    2056 is a palindrome when encoded as a pair of octets, so endianness does
    not affect it. Similarly, 257 is also 2-octet-palindromic. Meanwhile, 1 and
    256 are counterparts when encoded as pairs of octets. Because they are both
    valid values for day of year, it is impossible to make a conclusion about
    endianness based on day of year if it is either 1 or 256 in big-endian. On
    the other hand, 257 is a palindrome. These two facts combined means that we
    cannot determine the endianness of the header that starts on the
    aforementioned dates. The way this function was written, the endianness
    will be determined to be big in these cases. This problem is also recorded
    in libmseed.

    :param header: the fixed header of the data record
    :return: either of the string 'big' or 'little' depending on the extracted
    endianness of header
    """
    record_start_time_string = header.record_start_time
    record_start_time_tuple = struct.unpack('>hhbbbbh',
                                            record_start_time_string)
    # libmseed uses 1900 to 2100 as the sane year range. We follow their
    # example here.
    year_is_good = (1900 <= record_start_time_tuple[0] <= 2100)
    # The upper range is 366 to account for leap years.
    day_is_good = (1 <= record_start_time_tuple[1] <= 366)

    endianness = 'big' if year_is_good and day_is_good else 'little'
    return endianness


def get_data_endianness(blockette_1000: Blockette1000):
    """
    Get endianness of a data record by examining blockette 1000.

    :param fixed_header: the fixed header of the data record.
    :param blockette_1000: the blockette 1000 of the data record.
    """
    # The byte order is only one byte so using big or little endian does not
    # matter.
    blockette_1000_endianness = int.from_bytes(
        blockette_1000.byte_order, 'big'
    )
    if blockette_1000_endianness:
        return 'big'
    else:
        return 'little'


def calculate_sample_rate(factor: int, multiplier: int) -> float:
    """
    Calculate the sample rate using the sample rate factor and multiplier. This
    algorithm is described around the start of Chapter 8 in the SEED manual.

    :param factor: the sample rate factor
    :param multiplier: the sample rate multiplier
    :return: the nominal sample rate
    """
    sample_rate = 0
    if factor == 0:
        sample_rate = 0
    elif factor > 0 and multiplier > 0:
        sample_rate = factor * multiplier
    elif factor > 0 and multiplier < 0:
        sample_rate = -(factor / multiplier)
    elif factor < 0 and multiplier > 0:
        sample_rate = -(multiplier / factor)
    elif factor < 0 and multiplier < 0:
        sample_rate = 1 / (factor * multiplier)
    return sample_rate


def get_record_metadata(header: FixedHeader, header_unpacker: Unpacker):
    station = header.station.decode('utf-8').rstrip()
    location = header.location.decode('utf-8').rstrip()
    channel = header.channel.decode('utf-8').rstrip()
    network = header.net_code.decode('utf-8').rstrip()

    record_start_time_string = header.record_start_time
    record_start_time_tuple = header_unpacker.unpack('HHBBBBH',
                                                     record_start_time_string)
    record_start_time = UTCDateTime(year=record_start_time_tuple[0],
                                    julday=record_start_time_tuple[1],
                                    hour=record_start_time_tuple[2],
                                    minute=record_start_time_tuple[3],
                                    second=record_start_time_tuple[4],
                                    microsecond=record_start_time_tuple[
                                                    6] * 100)

    sample_count = header_unpacker.unpack('H', header.sample_count)[0]

    sample_rate_factor = header_unpacker.unpack(
        'h', header.sample_rate_factor
    )[0]
    sample_rate_multiplier = header_unpacker.unpack(
        'h', header.sample_rate_multiplier
    )[0]
    sample_rate = calculate_sample_rate(sample_rate_factor,
                                        sample_rate_multiplier)

    return RecordMetadata(station, location, channel, network,
                          record_start_time, sample_count, sample_rate)
